@@ -103,14 +103,6 @@ typedef struct
 static int transport_bio_simple_init(BIO* bio, SOCKET socket, int shutdown);
 static int transport_bio_simple_uninit(BIO* bio);
 
-static long transport_bio_simple_callback(WINPR_ATTR_UNUSED BIO* bio, WINPR_ATTR_UNUSED int mode,
-                                          WINPR_ATTR_UNUSED const char* argp,
-                                          WINPR_ATTR_UNUSED int argi, WINPR_ATTR_UNUSED long argl,
-                                          WINPR_ATTR_UNUSED long ret)
-{
-	return 1;
-}
-
 static int transport_bio_simple_write(BIO* bio, const char* buf, int size)
 {
 	int error = 0;
@@ -464,14 +456,6 @@ typedef struct
 	RingBuffer xmitBuffer;
 } WINPR_BIO_BUFFERED_SOCKET;
 
-static long transport_bio_buffered_callback(WINPR_ATTR_UNUSED BIO* bio, WINPR_ATTR_UNUSED int mode,
-                                            WINPR_ATTR_UNUSED const char* argp,
-                                            WINPR_ATTR_UNUSED int argi, WINPR_ATTR_UNUSED long argl,
-                                            WINPR_ATTR_UNUSED long ret)
-{
-	return 1;
-}
-
 static int transport_bio_buffered_write(BIO* bio, const char* buf, int num)
 {
 	int ret = num;
@@ -1370,29 +1354,17 @@ static int freerdp_tcp_layer_read(void* userContext, void* data, int bytes)
 
 	rdpTcpLayer* tcpLayer = (rdpTcpLayer*)userContext;
 
-	int error = 0;
-	int status = 0;
-
 	(void)WSAResetEvent(tcpLayer->hEvent);
-	status = _recv((SOCKET)tcpLayer->sockfd, data, bytes, 0);
+	const int status = _recv((SOCKET)tcpLayer->sockfd, data, bytes, 0);
 
 	if (status > 0)
 		return status;
 
-	if (status == 0)
-		return -1; /* socket closed */
-
-	error = WSAGetLastError();
+	const int error = WSAGetLastError();
 
 	if ((error == WSAEWOULDBLOCK) || (error == WSAEINTR) || (error == WSAEINPROGRESS) ||
 	    (error == WSAEALREADY))
-	{
-		status = 0;
-	}
-	else
-	{
-		status = -1;
-	}
+		errno = EAGAIN;
 
 	return status;
 }
@@ -1406,25 +1378,15 @@ static int freerdp_tcp_layer_write(void* userContext, const void* data, int byte
 
 	rdpTcpLayer* tcpLayer = (rdpTcpLayer*)userContext;
 
-	int error = 0;
-	int status = 0;
+	const int status = _send((SOCKET)tcpLayer->sockfd, data, bytes, 0);
+	if (status > 0)
+		return status;
 
-	status = _send((SOCKET)tcpLayer->sockfd, data, bytes, 0);
+	const int error = WSAGetLastError();
 
-	if (status <= 0)
-	{
-		error = WSAGetLastError();
-
-		if ((error == WSAEWOULDBLOCK) || (error == WSAEINTR) || (error == WSAEINPROGRESS) ||
-		    (error == WSAEALREADY))
-		{
-			status = 0;
-		}
-		else
-		{
-			status = -1;
-		}
-	}
+	if ((error == WSAEWOULDBLOCK) || (error == WSAEINTR) || (error == WSAEINPROGRESS) ||
+	    (error == WSAEALREADY))
+		errno = EAGAIN;
 
 	return status;
 }
