@@ -28,15 +28,18 @@
 #include <errno.h>
 
 #include <winpr/crt.h>
+#include <winpr/winpr.h>
 #include <winpr/string.h>
 #include <winpr/sspi.h>
 #include <winpr/ssl.h>
+#include <winpr/json.h>
 
 #include <winpr/stream.h>
 #include <freerdp/utils/ringbuffer.h>
 
 #include <freerdp/crypto/certificate.h>
 #include <freerdp/crypto/certificate_data.h>
+#include <freerdp/utils/helpers.h>
 
 #include <freerdp/log.h>
 #include "../crypto/tls.h"
@@ -161,6 +164,8 @@ static int bio_rdp_tls_write(BIO* bio, const char* buf, int size)
 			case SSL_ERROR_SSL:
 				BIO_clear_flags(bio, BIO_FLAGS_SHOULD_RETRY);
 				break;
+			default:
+				break;
 		}
 	}
 
@@ -184,6 +189,7 @@ static int bio_rdp_tls_read(BIO* bio, char* buf, int size)
 
 	if (status <= 0)
 	{
+
 		switch (error)
 		{
 			case SSL_ERROR_NONE:
@@ -224,6 +230,8 @@ static int bio_rdp_tls_read(BIO* bio, char* buf, int size)
 			case SSL_ERROR_SYSCALL:
 				BIO_clear_flags(bio, BIO_FLAGS_SHOULD_RETRY);
 				break;
+			default:
+				break;
 		}
 	}
 
@@ -240,19 +248,18 @@ static int bio_rdp_tls_read(BIO* bio, char* buf, int size)
 
 static int bio_rdp_tls_puts(BIO* bio, const char* str)
 {
-	size_t size = 0;
-	int status = 0;
-
 	if (!str)
 		return 0;
 
-	size = strlen(str);
+	const size_t size = strnlen(str, INT_MAX + 1UL);
+	if (size > INT_MAX)
+		return -1;
 	ERR_clear_error();
-	status = BIO_write(bio, str, size);
-	return status;
+	return BIO_write(bio, str, (int)size);
 }
 
-static int bio_rdp_tls_gets(BIO* bio, char* str, int size)
+static int bio_rdp_tls_gets(WINPR_ATTR_UNUSED BIO* bio, WINPR_ATTR_UNUSED char* str,
+                            WINPR_ATTR_UNUSED int size)
 {
 	return 1;
 }
@@ -262,7 +269,7 @@ static long bio_rdp_tls_ctrl(BIO* bio, int cmd, long num, void* ptr)
 	BIO* ssl_rbio = NULL;
 	BIO* ssl_wbio = NULL;
 	BIO* next_bio = NULL;
-	int status = -1;
+	long status = -1;
 	BIO_RDP_TLS* tls = (BIO_RDP_TLS*)BIO_get_data(bio);
 
 	if (!tls)
@@ -309,8 +316,16 @@ static long bio_rdp_tls_ctrl(BIO* bio, int cmd, long num, void* ptr)
 			break;
 
 		case BIO_CTRL_GET_CALLBACK:
-			*((ULONG_PTR*)ptr) = (ULONG_PTR)SSL_get_info_callback(tls->ssl);
-			status = 1;
+			/* The OpenSSL API is horrible here:
+			 * we get a function pointer returned and have to cast it to ULONG_PTR
+			 * to return the value to the caller.
+			 *
+			 * This, of course, is something compilers warn about. So silence it by casting */
+			{
+				void* vptr = WINPR_FUNC_PTR_CAST(SSL_get_info_callback(tls->ssl), void*);
+				*((void**)ptr) = vptr;
+				status = 1;
+			}
 			break;
 
 		case BIO_C_SSL_MODE:
@@ -439,7 +454,8 @@ static long bio_rdp_tls_ctrl(BIO* bio, int cmd, long num, void* ptr)
 
 			if (status <= 0)
 			{
-				switch (SSL_get_error(tls->ssl, status))
+				const int err = (status < INT32_MIN) ? INT32_MIN : (int)status;
+				switch (SSL_get_error(tls->ssl, err))
 				{
 					case SSL_ERROR_WANT_READ:
 						BIO_set_flags(bio, BIO_FLAGS_READ | BIO_FLAGS_SHOULD_RETRY);
@@ -517,12 +533,11 @@ static int bio_rdp_tls_free(BIO* bio)
 static long bio_rdp_tls_callback_ctrl(BIO* bio, int cmd, bio_info_cb* fp)
 {
 	long status = 0;
-	BIO_RDP_TLS* tls = NULL;
 
 	if (!bio)
 		return 0;
 
-	tls = (BIO_RDP_TLS*)BIO_get_data(bio);
+	BIO_RDP_TLS* tls = (BIO_RDP_TLS*)BIO_get_data(bio);
 
 	if (!tls)
 		return 0;
@@ -532,10 +547,12 @@ static long bio_rdp_tls_callback_ctrl(BIO* bio, int cmd, bio_info_cb* fp)
 		case BIO_CTRL_SET_CALLBACK:
 		{
 			typedef void (*fkt_t)(const SSL*, int, int);
+
 			/* Documented since https://www.openssl.org/docs/man1.1.1/man3/BIO_set_callback.html
 			 * the argument is not really of type bio_info_cb* and must be cast
 			 * to the required type */
-			fkt_t fkt = (fkt_t)(void*)fp;
+
+			fkt_t fkt = WINPR_FUNC_PTR_CAST(fp, fkt_t);
 			SSL_set_info_callback(tls->ssl, fkt);
 			status = 1;
 		}
@@ -633,9 +650,7 @@ static SecPkgContext_Bindings* tls_get_channel_bindings(const rdpCertificate* ce
 {
 	size_t CertificateHashLength = 0;
 	BYTE* ChannelBindingToken = NULL;
-	UINT32 ChannelBindingTokenLength = 0;
 	SEC_CHANNEL_BINDINGS* ChannelBindings = NULL;
-	SecPkgContext_Bindings* ContextBindings = NULL;
 	const size_t PrefixLength = strnlen(TLS_SERVER_END_POINT, ARRAYSIZE(TLS_SERVER_END_POINT));
 
 	WINPR_ASSERT(cert);
@@ -661,20 +676,24 @@ static SecPkgContext_Bindings* tls_get_channel_bindings(const rdpCertificate* ce
 	if (!CertificateHash)
 		return NULL;
 
-	ChannelBindingTokenLength = PrefixLength + CertificateHashLength;
-	ContextBindings = (SecPkgContext_Bindings*)calloc(1, sizeof(SecPkgContext_Bindings));
+	const size_t ChannelBindingTokenLength = PrefixLength + CertificateHashLength;
+	SecPkgContext_Bindings* ContextBindings = calloc(1, sizeof(SecPkgContext_Bindings));
 
 	if (!ContextBindings)
 		goto out_free;
 
-	ContextBindings->BindingsLength = sizeof(SEC_CHANNEL_BINDINGS) + ChannelBindingTokenLength;
+	const size_t slen = sizeof(SEC_CHANNEL_BINDINGS) + ChannelBindingTokenLength;
+	if (slen > UINT32_MAX)
+		goto out_free;
+
+	ContextBindings->BindingsLength = (UINT32)slen;
 	ChannelBindings = (SEC_CHANNEL_BINDINGS*)calloc(1, ContextBindings->BindingsLength);
 
 	if (!ChannelBindings)
 		goto out_free;
 
 	ContextBindings->Bindings = ChannelBindings;
-	ChannelBindings->cbApplicationDataLength = ChannelBindingTokenLength;
+	ChannelBindings->cbApplicationDataLength = (UINT32)ChannelBindingTokenLength;
 	ChannelBindings->dwApplicationDataOffset = sizeof(SEC_CHANNEL_BINDINGS);
 	ChannelBindingToken = &((BYTE*)ChannelBindings)[ChannelBindings->dwApplicationDataOffset];
 	memcpy(ChannelBindingToken, TLS_SERVER_END_POINT, PrefixLength);
@@ -690,7 +709,9 @@ out_free:
 static INIT_ONCE secrets_file_idx_once = INIT_ONCE_STATIC_INIT;
 static int secrets_file_idx = -1;
 
-static BOOL CALLBACK secrets_file_init_cb(PINIT_ONCE once, PVOID param, PVOID* context)
+static BOOL CALLBACK secrets_file_init_cb(WINPR_ATTR_UNUSED PINIT_ONCE once,
+                                          WINPR_ATTR_UNUSED PVOID param,
+                                          WINPR_ATTR_UNUSED PVOID* context)
 {
 	secrets_file_idx = SSL_get_ex_new_index(0, NULL, NULL, NULL, NULL);
 
@@ -728,7 +749,7 @@ static void tls_reset(rdpTls* tls)
 	}
 
 	/* tls->underlying is a stacked BIO under tls->bio.
-	 * BIO_free_all will free recursivly. */
+	 * BIO_free_all will free recursively. */
 	if (tls->bio)
 		BIO_free_all(tls->bio);
 	else if (tls->underlying)
@@ -765,7 +786,7 @@ static BOOL tls_prepare(rdpTls* tls, BIO* underlying, SSL_METHOD* method, int op
 	}
 
 	SSL_CTX_set_mode(tls->ctx, SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER | SSL_MODE_ENABLE_PARTIAL_WRITE);
-	SSL_CTX_set_options(tls->ctx, options);
+	SSL_CTX_set_options(tls->ctx, WINPR_ASSERTING_INT_CAST(uint64_t, options));
 	SSL_CTX_set_read_ahead(tls->ctx, 1);
 #if OPENSSL_VERSION_NUMBER >= 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
 	UINT16 version = freerdp_settings_get_uint16(settings, FreeRDP_TLSMinVersion);
@@ -782,7 +803,7 @@ static BOOL tls_prepare(rdpTls* tls, BIO* underlying, SSL_METHOD* method, int op
 	}
 #endif
 #if OPENSSL_VERSION_NUMBER >= 0x10100000L && !defined(LIBRESSL_VERSION_NUMBER)
-	SSL_CTX_set_security_level(tls->ctx, settings->TlsSecLevel);
+	SSL_CTX_set_security_level(tls->ctx, WINPR_ASSERTING_INT_CAST(int, settings->TlsSecLevel));
 #endif
 
 	if (settings->AllowedTlsCiphers)
@@ -821,7 +842,8 @@ static BOOL tls_prepare(rdpTls* tls, BIO* underlying, SSL_METHOD* method, int op
 	return TRUE;
 }
 
-static void adjustSslOptions(const int* options)
+static void
+adjustSslOptions(WINPR_ATTR_UNUSED int* options) // NOLINT(readability-non-const-parameter)
 {
 	WINPR_ASSERT(options);
 #if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
@@ -891,8 +913,10 @@ TlsHandshakeResult freerdp_tls_connect_ex(rdpTls* tls, BIO* underlying, const SS
 	if (!tls_prepare(tls, underlying, methods, options, TRUE))
 		return 0;
 
-#if !defined(OPENSSL_NO_TLSEXT) && !defined(LIBRESSL_VERSION_NUMBER)
-	SSL_set_tlsext_host_name(tls->ssl, tls_get_server_name(tls));
+#if !defined(OPENSSL_NO_TLSEXT)
+	const char* str = tls_get_server_name(tls);
+	void* ptr = WINPR_CAST_CONST_PTR_AWAY(str, void*);
+	SSL_set_tlsext_host_name(tls->ssl, ptr);
 #endif
 
 	return freerdp_tls_handshake(tls);
@@ -910,7 +934,7 @@ TlsHandshakeResult freerdp_tls_handshake(rdpTls* tls)
 	TlsHandshakeResult ret = TLS_HANDSHAKE_ERROR;
 
 	WINPR_ASSERT(tls);
-	int status = BIO_do_handshake(tls->bio);
+	const long status = BIO_do_handshake(tls->bio);
 	if (status != 1)
 	{
 		if (!BIO_should_retry(tls->bio))
@@ -956,7 +980,9 @@ TlsHandshakeResult freerdp_tls_handshake(rdpTls* tls)
 
 		if (tls->isClientMode)
 		{
-			verify_status = tls_verify_certificate(tls, cert, tls_get_server_name(tls), tls->port);
+			WINPR_ASSERT(tls->port <= UINT16_MAX);
+			verify_status =
+			    tls_verify_certificate(tls, cert, tls_get_server_name(tls), (UINT16)tls->port);
 
 			if (verify_status < 1)
 			{
@@ -977,25 +1003,28 @@ static int pollAndHandshake(rdpTls* tls)
 
 	do
 	{
-		HANDLE event = NULL;
+		HANDLE events[] = { freerdp_abort_event(tls->context), NULL };
 		DWORD status = 0;
-		if (BIO_get_event(tls->bio, &event) < 0)
+		if (BIO_get_event(tls->bio, &events[1]) < 0)
 		{
 			WLog_ERR(TAG, "unable to retrieve BIO associated event");
 			return -1;
 		}
 
-		if (!event)
+		if (!events[1])
 		{
 			WLog_ERR(TAG, "unable to retrieve BIO event");
 			return -1;
 		}
 
-		status = WaitForSingleObjectEx(event, 50, TRUE);
+		status = WaitForMultipleObjectsEx(ARRAYSIZE(events), events, FALSE, INFINITE, TRUE);
 		switch (status)
 		{
-			case WAIT_OBJECT_0:
+			case WAIT_OBJECT_0 + 1:
 				break;
+			case WAIT_OBJECT_0:
+				WLog_DBG(TAG, "Abort event set, cancel connect");
+				return -1;
 			case WAIT_TIMEOUT:
 			case WAIT_IO_COMPLETION:
 				continue;
@@ -1034,6 +1063,8 @@ int freerdp_tls_connect(rdpTls* tls, BIO* underlying)
 		case TLS_HANDSHAKE_ERROR:
 		case TLS_HANDSHAKE_VERIFY_ERROR:
 			return -1;
+		default:
+			return -1;
 	}
 
 	return pollAndHandshake(tls);
@@ -1077,7 +1108,7 @@ TlsHandshakeResult freerdp_tls_accept_ex(rdpTls* tls, BIO* underlying, rdpSettin
 {
 	WINPR_ASSERT(tls);
 
-	long options = 0;
+	int options = 0;
 	int status = 0;
 
 	/**
@@ -1171,6 +1202,19 @@ TlsHandshakeResult freerdp_tls_accept_ex(rdpTls* tls, BIO* underlying, rdpSettin
 		return TLS_HANDSHAKE_ERROR;
 	}
 
+	const size_t cnt = freerdp_certificate_get_chain_len(cert);
+	for (size_t x = 0; x < cnt; x++)
+	{
+		X509* xcert = freerdp_certificate_get_chain_at(cert, x);
+		WINPR_ASSERT(xcert);
+		const long rc = SSL_add1_chain_cert(tls->ssl, xcert);
+		if (rc != 1)
+		{
+			WLog_ERR(TAG, "SSL_add1_chain_cert failed");
+			return TLS_HANDSHAKE_ERROR;
+		}
+	}
+
 #if defined(MICROSOFT_IOS_SNI_BUG) && !defined(OPENSSL_NO_TLSEXT) && \
     !defined(LIBRESSL_VERSION_NUMBER)
 	SSL_set_tlsext_debug_callback(tls->ssl, tls_openssl_tlsext_debug_callback);
@@ -1189,10 +1233,10 @@ BOOL freerdp_tls_send_alert(rdpTls* tls)
 	if (!tls->ssl)
 		return TRUE;
 
-		/**
-		 * FIXME: The following code does not work on OpenSSL > 1.1.0 because the
-		 *        SSL struct is opaqe now
-		 */
+	/**
+	 * FIXME: The following code does not work on OpenSSL > 1.1.0 because the
+	 *        SSL struct is opaqe now
+	 */
 #if (!defined(LIBRESSL_VERSION_NUMBER) && (OPENSSL_VERSION_NUMBER < 0x10100000L)) || \
     (defined(LIBRESSL_VERSION_NUMBER) && (LIBRESSL_VERSION_NUMBER <= 0x2080300fL))
 
@@ -1226,40 +1270,41 @@ BOOL freerdp_tls_send_alert(rdpTls* tls)
 	return TRUE;
 }
 
-int freerdp_tls_write_all(rdpTls* tls, const BYTE* data, int length)
+int freerdp_tls_write_all(rdpTls* tls, const BYTE* data, size_t length)
 {
 	WINPR_ASSERT(tls);
-	int status = 0;
-	int offset = 0;
+	size_t offset = 0;
 	BIO* bio = tls->bio;
 
+	if (length > INT32_MAX)
+		return -1;
+
 	while (offset < length)
 	{
 		ERR_clear_error();
-		status = BIO_write(bio, &data[offset], length - offset);
+		const int status = BIO_write(bio, &data[offset], (int)(length - offset));
 
 		if (status > 0)
-		{
-			offset += status;
-		}
+			offset += (size_t)status;
 		else
 		{
 			if (!BIO_should_retry(bio))
 				return -1;
 
 			if (BIO_write_blocked(bio))
-				status = BIO_wait_write(bio, 100);
+			{
+				const long rc = BIO_wait_write(bio, 100);
+				if (rc < 0)
+					return -1;
+			}
 			else if (BIO_read_blocked(bio))
 				return -2; /* Abort write, there is data that must be read */
 			else
 				USleep(100);
-
-			if (status < 0)
-				return -1;
 		}
 	}
 
-	return length;
+	return (int)length;
 }
 
 int freerdp_tls_set_alert_code(rdpTls* tls, int level, int description)
@@ -1343,8 +1388,8 @@ static BOOL is_accepted(rdpTls* tls, const rdpCertificate* cert)
 			return TRUE;
 	}
 
-	freerdp_settings_set_string(settings, keyAccepted, NULL);
-	freerdp_settings_set_uint32(settings, keyLength, 0);
+	(void)freerdp_settings_set_string(settings, keyAccepted, NULL);
+	(void)freerdp_settings_set_uint32(settings, keyLength, 0);
 
 	return FALSE;
 }
@@ -1395,15 +1440,11 @@ static BOOL is_accepted_fingerprint(const rdpCertificate* cert,
 		{
 			char* subcontext = NULL;
 			const char* h = strtok_s(cur, ":", &subcontext);
-			const char* fp = NULL;
 
 			if (!h)
 				goto next;
 
-			fp = h + strlen(h) + 1;
-			if (!fp)
-				goto next;
-
+			const char* fp = h + strlen(h) + 1;
 			if (compare_fingerprint_all(fp, h, cert))
 			{
 				rc = TRUE;
@@ -1443,10 +1484,12 @@ static BOOL accept_cert(rdpTls* tls, const rdpCertificate* cert)
 
 	size_t pemLength = 0;
 	char* pem = freerdp_certificate_get_pem_ex(cert, &pemLength, FALSE);
-
 	BOOL rc = FALSE;
-	if (freerdp_settings_set_string_len(settings, id, pem, pemLength))
-		rc = freerdp_settings_set_uint32(settings, lid, pemLength);
+	if (pemLength <= UINT32_MAX)
+	{
+		if (freerdp_settings_set_string_len(settings, id, pem, pemLength))
+			rc = freerdp_settings_set_uint32(settings, lid, (UINT32)pemLength);
+	}
 	free(pem);
 	return rc;
 }
@@ -1460,6 +1503,131 @@ static BOOL tls_extract_full_pem(const rdpCertificate* cert, BYTE** PublicKey,
 	return *PublicKey != NULL;
 }
 
+static int tls_config_parse_bool(WINPR_JSON* json, const char* opt)
+{
+	WINPR_JSON* val = WINPR_JSON_GetObjectItem(json, opt);
+	if (!val || !WINPR_JSON_IsBool(val))
+		return -1;
+
+	if (WINPR_JSON_IsTrue(val))
+		return 1;
+	return 0;
+}
+
+static int tls_config_check_allowed_hashed(const char* configfile, const rdpCertificate* cert,
+                                           WINPR_JSON* json)
+{
+	WINPR_ASSERT(configfile);
+	WINPR_ASSERT(cert);
+	WINPR_ASSERT(json);
+
+	WINPR_JSON* db = WINPR_JSON_GetObjectItem(json, "certificate-db");
+	if (!db || !WINPR_JSON_IsArray(db))
+		return 0;
+
+	for (size_t x = 0; x < WINPR_JSON_GetArraySize(db); x++)
+	{
+		WINPR_JSON* cur = WINPR_JSON_GetArrayItem(db, x);
+		if (!cur || !WINPR_JSON_IsObject(cur))
+		{
+			WLog_WARN(TAG,
+			          "[%s] invalid certificate-db entry at position %" PRIuz ": not a JSON object",
+			          configfile, x);
+			continue;
+		}
+
+		WINPR_JSON* key = WINPR_JSON_GetObjectItem(cur, "type");
+		if (!key || !WINPR_JSON_IsString(key))
+		{
+			WLog_WARN(TAG,
+			          "[%s] invalid certificate-db entry at position %" PRIuz
+			          ": invalid 'type' element, expected type string",
+			          configfile, x);
+			continue;
+		}
+		WINPR_JSON* val = WINPR_JSON_GetObjectItem(cur, "hash");
+		if (!val || !WINPR_JSON_IsString(val))
+		{
+			WLog_WARN(TAG,
+			          "[%s] invalid certificate-db entry at position %" PRIuz
+			          ": invalid 'hash' element, expected type string",
+			          configfile, x);
+			continue;
+		}
+
+		const char* skey = WINPR_JSON_GetStringValue(key);
+		const char* sval = WINPR_JSON_GetStringValue(val);
+
+		char* hash = freerdp_certificate_get_fingerprint_by_hash_ex(cert, skey, FALSE);
+		if (!hash)
+		{
+			WLog_WARN(TAG,
+			          "[%s] invalid certificate-db entry at position %" PRIuz
+			          ": hash type '%s' not supported by certificate",
+			          configfile, x, skey);
+			continue;
+		}
+
+		const int cmp = _stricmp(hash, sval);
+		free(hash);
+
+		if (cmp == 0)
+			return 1;
+	}
+
+	return 0;
+}
+
+static int tls_config_check_certificate(const rdpCertificate* cert, BOOL* pAllowUserconfig)
+{
+	WINPR_ASSERT(cert);
+	WINPR_ASSERT(pAllowUserconfig);
+
+	int rc = 0;
+	const char configfile[] = "certificates.json";
+	WINPR_JSON* json = freerdp_GetJSONConfigFile(TRUE, configfile);
+
+	if (!json)
+	{
+		WLog_DBG(TAG, "No or no valid configuration file for certificate handling, asking user");
+		goto fail;
+	}
+
+	if (tls_config_parse_bool(json, "deny") > 0)
+	{
+		WLog_WARN(TAG, "[%s] certificate denied by configuration", configfile);
+		rc = -1;
+		goto fail;
+	}
+
+	if (tls_config_parse_bool(json, "ignore") > 0)
+	{
+		WLog_WARN(TAG, "[%s] certificate ignored by configuration", configfile);
+		rc = 1;
+		goto fail;
+	}
+
+	if (tls_config_check_allowed_hashed(configfile, cert, json) > 0)
+	{
+		WLog_WARN(TAG, "[%s] certificate manually accepted by configuration", configfile);
+		rc = 1;
+		goto fail;
+	}
+
+	if (tls_config_parse_bool(json, "deny-userconfig") > 0)
+	{
+		WLog_WARN(TAG, "[%s] configuration denies user to accept certificates", configfile);
+		rc = -1;
+		goto fail;
+	}
+
+fail:
+
+	*pAllowUserconfig = (rc == 0);
+	WINPR_JSON_Delete(json);
+	return rc;
+}
+
 int tls_verify_certificate(rdpTls* tls, const rdpCertificate* cert, const char* hostname,
                            UINT16 port)
 {
@@ -1476,12 +1644,11 @@ int tls_verify_certificate(rdpTls* tls, const rdpCertificate* cert, const char*
 	rdpCertificateData* certificate_data = NULL;
 	BYTE* pemCert = NULL;
 	DWORD flags = VERIFY_CERT_FLAG_NONE;
-	freerdp* instance = NULL;
 
 	WINPR_ASSERT(tls);
-	WINPR_ASSERT(tls->context->settings);
+	WINPR_ASSERT(tls->context);
 
-	instance = (freerdp*)tls->context->settings->instance;
+	freerdp* instance = tls->context->instance;
 	WINPR_ASSERT(instance);
 
 	if (freerdp_shall_disconnect_context(instance->context))
@@ -1531,8 +1698,14 @@ int tls_verify_certificate(rdpTls* tls, const rdpCertificate* cert, const char*
 		}
 	}
 	/* ignore certificate verification if user explicitly required it (discouraged) */
-	else if (tls->context->settings->IgnoreCertificate)
+	else if (freerdp_settings_get_bool(tls->context->settings, FreeRDP_IgnoreCertificate))
+	{
+		WLog_WARN(TAG, "[DANGER] Certificate not checked, /cert:ignore in use.");
+		WLog_WARN(TAG, "[DANGER] This prevents MITM attacks from being detected!");
+		WLog_WARN(TAG,
+		          "[DANGER] Avoid using this unless in a secure LAN (=no internet) environment");
 		verification_status = 1; /* success! */
+	}
 	else if (!tls->isGatewayTransport && (tls->context->settings->AuthenticationLevel == 0))
 		verification_status = 1; /* success! */
 	else
@@ -1582,9 +1755,13 @@ int tls_verify_certificate(rdpTls* tls, const rdpCertificate* cert, const char*
 		if (!hostname_match)
 			flags |= VERIFY_CERT_FLAG_MISMATCH;
 
+		BOOL allowUserconfig = TRUE;
+		if (!certificate_status || !hostname_match)
+			verification_status = tls_config_check_certificate(cert, &allowUserconfig);
+
 		/* verification could not succeed with OpenSSL, use known_hosts file and prompt user for
 		 * manual verification */
-		if (!certificate_status || !hostname_match)
+		if (allowUserconfig && (!certificate_status || !hostname_match))
 		{
 			DWORD accept_certificate = 0;
 			size_t pem_length = 0;
@@ -1814,7 +1991,8 @@ void tls_print_new_certificate_warn(rdpCertificateStore* store, const char* host
 	free(path);
 }
 
-void tls_print_certificate_error(rdpCertificateStore* store, rdpCertificateData* stored_data,
+void tls_print_certificate_error(rdpCertificateStore* store,
+                                 WINPR_ATTR_UNUSED rdpCertificateData* stored_data,
                                  const char* hostname, UINT16 port, const char* fingerprint)
 {
 	char* path = freerdp_certificate_store_get_cert_path(store, hostname, port);
