@@ -70,6 +70,12 @@ static const char* pf_modules_get_filter_type_string(PF_FILTER_TYPE result)
 			return "FILTER_TYPE_SERVER_PEER_LOGON";
 		case FILTER_TYPE_CLIENT_PASSTHROUGH_CHANNEL_CREATE:
 			return "FILTER_TYPE_CLIENT_PASSTHROUGH_CHANNEL_CREATE";
+		case FILTER_TYPE_STATIC_INTERCEPT_LIST:
+			return "FILTER_TYPE_STATIC_INTERCEPT_LIST";
+		case FILTER_TYPE_DYN_INTERCEPT_LIST:
+			return "FILTER_TYPE_DYN_INTERCEPT_LIST";
+		case FILTER_TYPE_INTERCEPT_CHANNEL:
+			return "FILTER_TYPE_INTERCEPT_CHANNEL";
 		case FILTER_LAST:
 			return "FILTER_LAST";
 		default:
@@ -332,8 +338,8 @@ BOOL pf_modules_run_filter(proxyModule* module, PF_FILTER_TYPE type, proxyData*
  * @context: current session server's rdpContext instance.
  * @info: pointer to per-session data.
  */
-static BOOL pf_modules_set_plugin_data(proxyPluginsManager* mgr, const char* plugin_name,
-                                       proxyData* pdata, void* data)
+static BOOL pf_modules_set_plugin_data(WINPR_ATTR_UNUSED proxyPluginsManager* mgr,
+                                       const char* plugin_name, proxyData* pdata, void* data)
 {
 	union
 	{
@@ -363,8 +369,8 @@ static BOOL pf_modules_set_plugin_data(proxyPluginsManager* mgr, const char* plu
  * if there's no data related to `plugin_name` in `context` (current session), a NULL will be
  * returned.
  */
-static void* pf_modules_get_plugin_data(proxyPluginsManager* mgr, const char* plugin_name,
-                                        proxyData* pdata)
+static void* pf_modules_get_plugin_data(WINPR_ATTR_UNUSED proxyPluginsManager* mgr,
+                                        const char* plugin_name, proxyData* pdata)
 {
 	union
 	{
@@ -378,7 +384,7 @@ static void* pf_modules_get_plugin_data(proxyPluginsManager* mgr, const char* pl
 	return HashTable_GetItemValue(pdata->modules_info, ccharconv.cp);
 }
 
-static void pf_modules_abort_connect(proxyPluginsManager* mgr, proxyData* pdata)
+static void pf_modules_abort_connect(WINPR_ATTR_UNUSED proxyPluginsManager* mgr, proxyData* pdata)
 {
 	WINPR_ASSERT(pdata);
 	WLog_DBG(TAG, "is called!");
@@ -479,13 +485,54 @@ void pf_modules_list_loaded_plugins(proxyModule* module)
 	ArrayList_ForEach(module->plugins, pf_modules_print_ArrayList_ForEachFkt);
 }
 
-static BOOL pf_modules_load_module(const char* module_path, proxyModule* module, void* userdata)
+static BOOL pf_modules_load_static_module(const char* module_name, proxyModule* module,
+                                          void* userdata)
 {
-	HMODULE handle = NULL;
-	proxyModuleEntryPoint pEntryPoint = NULL;
 	WINPR_ASSERT(module);
 
-	handle = LoadLibraryX(module_path);
+	HANDLE handle = GetModuleHandleA(NULL);
+
+	if (handle == NULL)
+	{
+		WLog_ERR(TAG, "failed loading static library: %s", module_name);
+		return FALSE;
+	}
+
+	char name[256] = { 0 };
+	(void)_snprintf(name, sizeof(name), "%s_%s", module_name, MODULE_ENTRY_POINT);
+	for (size_t x = 0; x < strnlen(name, sizeof(name)); x++)
+	{
+		if (name[x] == '-')
+			name[x] = '_';
+	}
+
+	proxyModuleEntryPoint pEntryPoint = GetProcAddressAs(handle, name, proxyModuleEntryPoint);
+	if (!pEntryPoint)
+	{
+		WLog_ERR(TAG, "GetProcAddress failed for static %s (module %s)", name, module_name);
+		goto error;
+	}
+	if (!ArrayList_Append(module->handles, handle))
+	{
+		WLog_ERR(TAG, "ArrayList_Append failed!");
+		goto error;
+	}
+	return pf_modules_add(module, pEntryPoint, userdata);
+
+error:
+	FreeLibrary(handle);
+	return FALSE;
+}
+
+static BOOL pf_modules_load_module(const char* module_path, const char* module_name,
+                                   proxyModule* module, void* userdata)
+{
+	WINPR_ASSERT(module);
+
+	if (pf_modules_load_static_module(module_name, module, userdata))
+		return TRUE;
+
+	HANDLE handle = LoadLibraryX(module_path);
 
 	if (handle == NULL)
 	{
@@ -493,7 +540,8 @@ static BOOL pf_modules_load_module(const char* module_path, proxyModule* module,
 		return FALSE;
 	}
 
-	pEntryPoint = (proxyModuleEntryPoint)GetProcAddress(handle, MODULE_ENTRY_POINT);
+	proxyModuleEntryPoint pEntryPoint =
+	    GetProcAddressAs(handle, MODULE_ENTRY_POINT, proxyModuleEntryPoint);
 	if (!pEntryPoint)
 	{
 		WLog_ERR(TAG, "GetProcAddress failed while loading %s", module_path);
@@ -586,7 +634,6 @@ proxyModule* pf_modules_new(const char* root_dir, const char** modules, size_t c
 			if (!winpr_PathMakePath(path, NULL))
 			{
 				WLog_ERR(TAG, "error occurred while creating modules directory: %s", root_dir);
-				goto error;
 			}
 		}
 
@@ -600,7 +647,7 @@ proxyModule* pf_modules_new(const char* root_dir, const char** modules, size_t c
 			(void)_snprintf(name, sizeof(name), "proxy-%s-plugin%s", modules[i],
 			                FREERDP_SHARED_LIBRARY_SUFFIX);
 			fullpath = GetCombinedPath(path, name);
-			pf_modules_load_module(fullpath, module, NULL);
+			pf_modules_load_module(fullpath, modules[i], module, NULL);
 			free(fullpath);
 		}
 	}
