@@ -96,14 +96,15 @@ static const WinPrAsn1_OID negoex_OID = { 10, (BYTE*)"\x2b\x06\x01\x04\x01\x82\x
 
 #ifdef WITH_KRB5
 static const SecPkg SecPkgTable[] = {
+	{ KERBEROS_SSP_NAME, &KERBEROS_SecurityFunctionTableA, &KERBEROS_SecurityFunctionTableW },
 	{ KERBEROS_SSP_NAME, &KERBEROS_SecurityFunctionTableA, &KERBEROS_SecurityFunctionTableW },
 	{ NTLM_SSP_NAME, &NTLM_SecurityFunctionTableA, &NTLM_SecurityFunctionTableW }
 };
 
 static const Mech MechTable[] = {
 	{ &kerberos_u2u_OID, &SecPkgTable[0], ISC_REQ_INTEGRITY | ISC_REQ_USE_SESSION_KEY, TRUE },
-	{ &kerberos_OID, &SecPkgTable[0], ISC_REQ_INTEGRITY, TRUE },
-	{ &ntlm_OID, &SecPkgTable[1], 0, FALSE },
+	{ &kerberos_OID, &SecPkgTable[1], ISC_REQ_INTEGRITY, TRUE },
+	{ &ntlm_OID, &SecPkgTable[2], 0, FALSE },
 };
 #else
 static const SecPkg SecPkgTable[] = { { NTLM_SSP_NAME, &NTLM_SecurityFunctionTableA,
@@ -248,7 +249,8 @@ static BOOL negotiate_get_dword(HKEY hKey, const char* subkey, DWORD* pdwValue)
 	return TRUE;
 }
 
-static BOOL negotiate_get_config_from_auth_package_list(void* pAuthData, BOOL* kerberos, BOOL* ntlm)
+static BOOL negotiate_get_config_from_auth_package_list(void* pAuthData, BOOL* kerberos, BOOL* ntlm,
+                                                        BOOL* u2u)
 {
 	char* tok_ctx = NULL;
 	char* tok_ptr = NULL;
@@ -270,14 +272,18 @@ static BOOL negotiate_get_config_from_auth_package_list(void* pAuthData, BOOL* k
 			PackageInclude = FALSE;
 		}
 
-		if (!_stricmp(PackageName, "ntlm"))
+		if (_stricmp(PackageName, "ntlm") == 0)
 		{
 			*ntlm = PackageInclude;
 		}
-		else if (!_stricmp(PackageName, "kerberos"))
+		else if (_stricmp(PackageName, "kerberos") == 0)
 		{
 			*kerberos = PackageInclude;
 		}
+		else if (_stricmp(PackageName, "u2u") == 0)
+		{
+			*u2u = PackageInclude;
+		}
 		else
 		{
 			WLog_WARN(TAG, "Unknown authentication package name: %s", PackageName);
@@ -290,22 +296,29 @@ static BOOL negotiate_get_config_from_auth_package_list(void* pAuthData, BOOL* k
 	return TRUE;
 }
 
-static BOOL negotiate_get_config(void* pAuthData, BOOL* kerberos, BOOL* ntlm)
+static BOOL negotiate_get_config(void* pAuthData, BOOL* kerberos, BOOL* ntlm, BOOL* u2u)
 {
 	HKEY hKey = NULL;
 	LONG rc = 0;
 
 	WINPR_ASSERT(kerberos);
 	WINPR_ASSERT(ntlm);
+	WINPR_ASSERT(u2u);
 
 #if !defined(WITH_KRB5_NO_NTLM_FALLBACK)
 	*ntlm = TRUE;
 #else
 	*ntlm = FALSE;
 #endif
+#if defined(WITH_KRB5)
 	*kerberos = TRUE;
+	*u2u = TRUE;
+#else
+	*kerberos = FALSE;
+	*u2u = FALSE;
+#endif
 
-	if (negotiate_get_config_from_auth_package_list(pAuthData, kerberos, ntlm))
+	if (negotiate_get_config_from_auth_package_list(pAuthData, kerberos, ntlm, u2u))
 	{
 		return TRUE; // use explicit authentication package list
 	}
@@ -417,11 +430,14 @@ static BOOL negotiate_write_neg_token(PSecBuffer output_buffer, NegToken* token)
 	if (!WinPrAsn1EncStreamSize(enc, &len) || len > output_buffer->cbBuffer)
 		goto cleanup;
 
+	if (len > UINT32_MAX)
+		goto cleanup;
+
 	Stream_StaticInit(&s, output_buffer->pvBuffer, len);
 
 	if (WinPrAsn1EncToStream(enc, &s))
 	{
-		output_buffer->cbBuffer = len;
+		output_buffer->cbBuffer = (UINT32)len;
 		ret = TRUE;
 	}
 
@@ -490,7 +506,10 @@ static BOOL negotiate_read_neg_token(PSecBuffer input, NegToken* token)
 					/* mechTypes [0] MechTypeList */
 					wStream s = WinPrAsn1DecGetStream(&dec2);
 					token->mechTypes.BufferType = SECBUFFER_TOKEN;
-					token->mechTypes.cbBuffer = Stream_Length(&s);
+					const size_t mlen = Stream_Length(&s);
+					if (mlen > UINT32_MAX)
+						return FALSE;
+					token->mechTypes.cbBuffer = (UINT32)mlen;
 					token->mechTypes.pvBuffer = Stream_Buffer(&s);
 					WLog_DBG(TAG, "\tmechTypes [0] (%li bytes)", token->mechTypes.cbBuffer);
 				}
@@ -525,7 +544,9 @@ static BOOL negotiate_read_neg_token(PSecBuffer input, NegToken* token)
 				/* mechToken [2] OCTET STRING */
 				if (!WinPrAsn1DecReadOctetString(&dec2, &octet_string, FALSE))
 					return FALSE;
-				token->mechToken.cbBuffer = octet_string.len;
+				if (octet_string.len > UINT32_MAX)
+					return FALSE;
+				token->mechToken.cbBuffer = (UINT32)octet_string.len;
 				token->mechToken.pvBuffer = octet_string.data;
 				token->mechToken.BufferType = SECBUFFER_TOKEN;
 				WLog_DBG(TAG, "\tmechToken [2] (%li bytes)", octet_string.len);
@@ -534,7 +555,9 @@ static BOOL negotiate_read_neg_token(PSecBuffer input, NegToken* token)
 				/* mechListMic [3] OCTET STRING */
 				if (!WinPrAsn1DecReadOctetString(&dec2, &octet_string, FALSE))
 					return FALSE;
-				token->mic.cbBuffer = octet_string.len;
+				if (octet_string.len > UINT32_MAX)
+					return FALSE;
+				token->mic.cbBuffer = (UINT32)octet_string.len;
 				token->mic.pvBuffer = octet_string.data;
 				token->mic.BufferType = SECBUFFER_TOKEN;
 				WLog_DBG(TAG, "\tmechListMIC [3] (%li bytes)", octet_string.len);
@@ -599,7 +622,11 @@ static SECURITY_STATUS negotiate_mic_exchange(NEGOTIATE_CONTEXT* context, NegTok
 	}
 
 	/* When using NTLM cipher states need to be reset after mic exchange */
-	if (_tcscmp(sspi_SecureHandleGetUpperPointer(&context->sub_context), NTLM_SSP_NAME) == 0)
+	const TCHAR* name = sspi_SecureHandleGetUpperPointer(&context->sub_context);
+	if (!name)
+		return SEC_E_INTERNAL_ERROR;
+
+	if (_tcsncmp(name, NTLM_SSP_NAME, ARRAYSIZE(NTLM_SSP_NAME)) == 0)
 	{
 		if (!ntlm_reset_cipher_state(&context->sub_context))
 			return SEC_E_INTERNAL_ERROR;
@@ -674,6 +701,9 @@ static SECURITY_STATUS SEC_ENTRY negotiate_InitializeSecurityContextW(
 			if (!init_context.mech)
 			{
 				/* Use the output buffer to store the optimistic token */
+				if (!output_buffer)
+					goto cleanup;
+
 				CopyMemory(&output_token.mechToken, output_buffer, sizeof(SecBuffer));
 
 				if (bindings_buffer)
@@ -720,7 +750,9 @@ static SECURITY_STATUS SEC_ENTRY negotiate_InitializeSecurityContextW(
 		{
 			init_context.spnego = TRUE;
 			init_context.mechTypes.BufferType = SECBUFFER_DATA;
-			init_context.mechTypes.cbBuffer = WinPrAsn1EncEndContainer(enc);
+			const size_t cb = WinPrAsn1EncEndContainer(enc);
+			WINPR_ASSERT(cb <= UINT32_MAX);
+			init_context.mechTypes.cbBuffer = (UINT32)cb;
 		}
 
 		/* Allocate memory for the new context */
@@ -771,7 +803,7 @@ static SECURITY_STATUS SEC_ENTRY negotiate_InitializeSecurityContextW(
 		if (!negotiate_read_neg_token(input_buffer, &input_token))
 			return SEC_E_INVALID_TOKEN;
 
-		/* On first response check if the server doesn't like out prefered mech */
+		/* On first response check if the server doesn't like out preferred mech */
 		if (context->state < NEGOTIATE_STATE_NEGORESP && input_token.supportedMech.len &&
 		    !sspi_gss_oid_compare(&input_token.supportedMech, context->mech->oid))
 		{
@@ -814,6 +846,8 @@ static SECURITY_STATUS SEC_ENTRY negotiate_InitializeSecurityContextW(
 					else
 						context->state = NEGOTIATE_STATE_FINAL;
 					break;
+				default:
+					break;
 			}
 
 			WLog_DBG(TAG, "Negotiated mechanism: %s", negotiate_mech_name(context->mech->oid));
@@ -822,6 +856,8 @@ static SECURITY_STATUS SEC_ENTRY negotiate_InitializeSecurityContextW(
 		if (context->state == NEGOTIATE_STATE_NEGORESP)
 		{
 			/* Store the mech token in the output buffer */
+			if (!output_buffer)
+				goto cleanup;
 			CopyMemory(&output_token.mechToken, output_buffer, sizeof(SecBuffer));
 
 			mech_input_buffers[0] = input_token.mechToken;
@@ -912,11 +948,12 @@ static const Mech* guessMech(PSecBuffer input_buffer, BOOL* spNego, WinPrAsn1_OI
 	WinPrAsn1Decoder decoder;
 	WinPrAsn1Decoder appDecoder;
 	WinPrAsn1_tagId tag = 0;
+	const char ssp[] = "NTLMSSP";
 
 	*spNego = FALSE;
 
 	/* Check for NTLM token */
-	if (input_buffer->cbBuffer >= 8 && strncmp(input_buffer->pvBuffer, "NTLMSSP", 8) == 0)
+	if (input_buffer->cbBuffer >= 8 && strncmp(input_buffer->pvBuffer, ssp, sizeof(ssp)) == 0)
 	{
 		*oid = ntlm_OID;
 		return negotiate_GetMechByOID(&ntlm_OID);
@@ -1127,7 +1164,8 @@ static SECURITY_STATUS SEC_ENTRY negotiate_AcceptSecurityContext(
 				return SEC_E_INVALID_TOKEN;
 
 			/* Use the output buffer to store the optimistic token */
-			CopyMemory(&output_token.mechToken, output_buffer, sizeof(SecBuffer));
+			if (output_buffer)
+				CopyMemory(&output_token.mechToken, output_buffer, sizeof(SecBuffer));
 
 			status = context->mech->pkg->table->AcceptSecurityContext(
 			    sub_cred, &context->sub_context, &mech_input, fContextReq | context->mech->flags,
@@ -1218,12 +1256,14 @@ static SECURITY_STATUS SEC_ENTRY negotiate_DeleteSecurityContext(PCtxtHandle phC
 	return status;
 }
 
-static SECURITY_STATUS SEC_ENTRY negotiate_ImpersonateSecurityContext(PCtxtHandle phContext)
+static SECURITY_STATUS SEC_ENTRY
+negotiate_ImpersonateSecurityContext(WINPR_ATTR_UNUSED PCtxtHandle phContext)
 {
 	return SEC_E_OK;
 }
 
-static SECURITY_STATUS SEC_ENTRY negotiate_RevertSecurityContext(PCtxtHandle phContext)
+static SECURITY_STATUS SEC_ENTRY
+negotiate_RevertSecurityContext(WINPR_ATTR_UNUSED PCtxtHandle phContext)
 {
 	return SEC_E_OK;
 }
@@ -1378,10 +1418,11 @@ static SECURITY_STATUS SEC_ENTRY negotiate_AcquireCredentialsHandleW(
     void* pAuthData, SEC_GET_KEY_FN pGetKeyFn, void* pvGetKeyArgument, PCredHandle phCredential,
     PTimeStamp ptsExpiry)
 {
-	BOOL kerberos = 0;
-	BOOL ntlm = 0;
+	BOOL kerberos = FALSE;
+	BOOL ntlm = FALSE;
+	BOOL u2u = FALSE;
 
-	if (!negotiate_get_config(pAuthData, &kerberos, &ntlm))
+	if (!negotiate_get_config(pAuthData, &kerberos, &ntlm, &u2u))
 		return SEC_E_INTERNAL_ERROR;
 
 	MechCred* creds = calloc(MECH_COUNT, sizeof(MechCred));
@@ -1395,9 +1436,11 @@ static SECURITY_STATUS SEC_ENTRY negotiate_AcquireCredentialsHandleW(
 		const SecPkg* pkg = MechTable[i].pkg;
 		cred->mech = &MechTable[i];
 
-		if (!kerberos && _tcscmp(pkg->name, KERBEROS_SSP_NAME) == 0)
+		if (!kerberos && sspi_gss_oid_compare(MechTable[i].oid, &kerberos_OID))
 			continue;
-		if (!ntlm && _tcscmp(SecPkgTable[i].name, NTLM_SSP_NAME) == 0)
+		if (!u2u && sspi_gss_oid_compare(MechTable[i].oid, &kerberos_u2u_OID))
+			continue;
+		if (!ntlm && _tcsncmp(SecPkgTable[i].name, NTLM_SSP_NAME, ARRAYSIZE(NTLM_SSP_NAME)) == 0)
 			continue;
 
 		WINPR_ASSERT(pkg->table_w);
@@ -1420,10 +1463,11 @@ static SECURITY_STATUS SEC_ENTRY negotiate_AcquireCredentialsHandleA(
     void* pAuthData, SEC_GET_KEY_FN pGetKeyFn, void* pvGetKeyArgument, PCredHandle phCredential,
     PTimeStamp ptsExpiry)
 {
-	BOOL kerberos = 0;
-	BOOL ntlm = 0;
+	BOOL kerberos = FALSE;
+	BOOL ntlm = FALSE;
+	BOOL u2u = FALSE;
 
-	if (!negotiate_get_config(pAuthData, &kerberos, &ntlm))
+	if (!negotiate_get_config(pAuthData, &kerberos, &ntlm, &u2u))
 		return SEC_E_INTERNAL_ERROR;
 
 	MechCred* creds = calloc(MECH_COUNT, sizeof(MechCred));
@@ -1438,9 +1482,11 @@ static SECURITY_STATUS SEC_ENTRY negotiate_AcquireCredentialsHandleA(
 
 		cred->mech = &MechTable[i];
 
-		if (!kerberos && _tcscmp(pkg->name, KERBEROS_SSP_NAME) == 0)
+		if (!kerberos && sspi_gss_oid_compare(MechTable[i].oid, &kerberos_OID))
 			continue;
-		if (!ntlm && _tcscmp(SecPkgTable[i].name, NTLM_SSP_NAME) == 0)
+		if (!u2u && sspi_gss_oid_compare(MechTable[i].oid, &kerberos_u2u_OID))
+			continue;
+		if (!ntlm && _tcsncmp(SecPkgTable[i].name, NTLM_SSP_NAME, ARRAYSIZE(NTLM_SSP_NAME)) == 0)
 			continue;
 
 		WINPR_ASSERT(pkg->table);
@@ -1458,17 +1504,17 @@ static SECURITY_STATUS SEC_ENTRY negotiate_AcquireCredentialsHandleA(
 	return SEC_E_OK;
 }
 
-static SECURITY_STATUS SEC_ENTRY negotiate_QueryCredentialsAttributesW(PCredHandle phCredential,
-                                                                       ULONG ulAttribute,
-                                                                       void* pBuffer)
+static SECURITY_STATUS SEC_ENTRY negotiate_QueryCredentialsAttributesW(
+    WINPR_ATTR_UNUSED PCredHandle phCredential, WINPR_ATTR_UNUSED ULONG ulAttribute,
+    WINPR_ATTR_UNUSED void* pBuffer)
 {
 	WLog_ERR(TAG, "TODO: Implement");
 	return SEC_E_UNSUPPORTED_FUNCTION;
 }
 
-static SECURITY_STATUS SEC_ENTRY negotiate_QueryCredentialsAttributesA(PCredHandle phCredential,
-                                                                       ULONG ulAttribute,
-                                                                       void* pBuffer)
+static SECURITY_STATUS SEC_ENTRY negotiate_QueryCredentialsAttributesA(
+    WINPR_ATTR_UNUSED PCredHandle phCredential, WINPR_ATTR_UNUSED ULONG ulAttribute,
+    WINPR_ATTR_UNUSED void* pBuffer)
 {
 	WLog_ERR(TAG, "TODO: Implement");
 	return SEC_E_UNSUPPORTED_FUNCTION;
